--                        -*- Mode: lua; lua-indent-level: 2; tab-width: 2 -*-
--[[
LibFilter_Outfitter

Author: Daphne Pfister <tiiroib02@sneakemail.com>

Thanks to John Stephen for writing Outfitter

License

Copyright (c) 2008 Daphne Pfister

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

-- I tried to follow the variable naming convention used by Outfitter
-- to allow this code to be included in that addon if the author wishes.

]]

local MAJOR, MINOR = "LibFilter-Slot-Outfitter", 303 -- tonumber(("$Revisions$):match("(%d+)"))
local filter = LibStub:NewLibrary(MAJOR, MINOR)
if not filter then return end

-- Avoid Global Lookups
local _G          = getfenv(0)
local GetItemInfo = _G.GetItemInfo
local ipairs      = _G.ipairs
local next        = _G.next
local pairs       = _G.pairs
local select      = _G.select
local string      = _G.string
local strjoin     = _G.strjoin
local table       = _G.table
local type        = _G.type
local unpack      = _G.unpack

local strmatch    = string.match
local tinsert     = table.insert
local tsort       = table.sort

-- Some icons
local ICON_EMPTY_CRATE = "Interface\\Icons\\INV_Crate_07"
local ICON_OUTFITTER   = "Interface\\Addons\\Outfitter\\Textures\\Icon"
local ICON_QUESTION    = "Interface\\Icons\\INV_Misc_QuestionMark"
local ICON_WEAPONS     = "Interface\\Icons\\Achievement_Arena_5v5_3"

-- No need for a full blown locale library here
local L
do
	local LL = {
		enUS = {
			["Outfitter Set:%s"] = true,
			["Unknown"] = true,
			["Outfitter Sets"] = true,
			["Set Name"] = true,
			["Any"] = true,
			["None"] = true,
			["Outfitter: |cFFFF0000no sets defined|r"] = true,
			["Outfitter: |cFFFF0000not loaded|r"] = true,
			["Outfitter"] = true,
			["Filter by Outfitter set"] = true,
			["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = true,
		},
		koKR = {
			["Outfitter Set:%s"] = "Outfitter 세트명:%s",
			["Unknown"] = "가격미확인", -- check
			["Outfitter Sets"] = "Outfitter 세트",
			["Set Name"] = "세트명",
			["Any"] = "모두",
			["None"] = "없음", -- check
			--["Outfitter: |cFFFF0000no sets defined|r"] = true, -- fix
			--["Outfitter: |cFFFF0000not loaded|r"] = true, -- fix
			--["Outfitter"] = true,
			["Filter by Outfitter set"] = "Outfitter 세트에 따라 필터링",
			["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = "마법부여나 키트 효과등을 무시하고 아이템 ID에 따라 분류합니다",
		},
		-- Based on swix's translation for ClosetGnome
		roRU = {
			["Outfitter Set:%s"] = "Набор Outfitter:%s",
			["Unknown"] = "Неизвестно",
			["Outfitter Sets"] = "Наборы Outfitter",
			["Any"] = "Любой",
			["None"] = "Ни один",
			--["Outfitter: |cFFFF0000no sets defined|r"] = true, -- fix
			--["Outfitter: |cFFFF0000not loaded|r"] = true, -- fix
			["Outfitter"] = "Outfitter",
			["Filter by Outfitter set"] = "Фильтр по наборам Outfitter",
			["Match by the item number, ignoring enchantments, variations, armor kits, etc."] = "Соответствие по номеру вещи, игнорирование зачарований, вариаций, коплектов брони и т.п.",
		},
	}

	local gGameLocale = GetLocale()
	local enL = LL["enUS"]
	L = LL[gGameLocale] or LL["enUS"]

	for vKey,vValue in pairs(enL) do
		if not L[vKey] then L[vKey] = vValue end
	end
	for vKey,vValue in pairs(L) do
		if vValue == true then L[vKey] = vKey end
	end
end

filter.author = "Nayala"
filter.credits = "Outfitter was written by John Stephen"
filter.description = L["Filter by Outfitter set"]
filter.filtertype = "Slot"
filter.filtertypes = filter.filtertypes or {}
filter.filtertypes["item"] = 1
filter.filtertypes["Slot"] = 1
filter.key = "Outfitter"
filter.name = L["Outfitter"]
filter.icon = ICON_OUTFITTER

filter.callbacks = filter.callbacks or LibStub("CallbackHandler-1.0"):New(filter)
filter.events = filter.events or {}
filter.myFrame = filter.myFrame or CreateFrame("Frame","LibFilter_Outfitter_Frame")
local callbacks = filter.callbacks
local events = filter.events
local myFrame = filter.myFrame

local gSetsTable = nil
local gSetIcons = nil

-- Prototypes
local CacheSets -- ()
local UpgradeSettings -- (pSettings)
local ExpandSettings -- (pSettings)
local NormalizeSettings -- (pSettings)
local Update
local HookOutfitterOutfitEvent -- (pEvent, pTarget)
local HookOutfitter
local UpdateNextFrame
local ToggleSettingsSet -- (pSettings, pSet)

function CacheSets()
	if not Outfitter.Initialized then return end

	gSetsTable = {}
	gSetIcons = {}

	local vCategoryOrder = Outfitter:GetCategoryOrder()
	for vCategoryIndex, vCategoryID in ipairs(vCategoryOrder) do
		local vOutfits = Outfitter:GetOutfitsByCategoryID(vCategoryID)

		for vIndex, vOutfit in ipairs(vOutfits) do
			if vOutfit and vOutfit.Name then
				gSetsTable[vOutfit.Name] = vOutfit
				if Outfitter.OutfitBar then
					gSetIcons[vOutfit.Name] = Outfitter.OutfitBar:GetOutfitTexture(vOutfit)
				end
			end
		end
	end
end

function UpgradeSettings(pSettings)
	-- No real upgrade path from Thorias's Baggins_Outfitter here
	-- Any rules from that pluggin will be treated as match any.
	local vSets = pSettings.sets
	if not vSets then
		local vSet = pSettings.set
		if not vSet or vSet == "Any" then
			vSets = { Any = true }
		elseif vSet == "None" then
			vSets = { }
		else
			vSets = { [vSet] = true }
		end
		pSettings.sets = vSets
	end
	return vSets
end

function ExpandSettings(pSettings)
	local vSets = pSettings.sets or UpgradeSettings(pSettings)

	if vSets.Any and gSetsTable then
		for vSet in pairs(gSetsTable) do
			vSets[vSet] = true
		end
	end
	return vSets
end

function NormalizeSettings(pSettings)
	local vSets = pSettings.sets or UpgradeSettings(pSettings)

	local vNextSet = next(vSets)

	-- No sets selected
	if not vNextSet then
		pSettings.set = "None"
		return vSets
	end

	-- Just one set selected
	if not next(vSets,vNextSet) then
		pSettings.set = vNextSet
		return vSets
	end

	-- Okay check if all sets are selected
	if gSetsTable then
		vSets.Any = true
		for vSet in pairs(gSetsTable) do
			if not vSets[vSet] then vSets.Any = nil end
		end
	end

	if vSets.Any then
		vSets = { Any = true }
		pSettings.sets = vSets
		pSettings.set = "Any"
		return vSets
	end

	pSettings.set = "????"
	return vSets
end

function Update()
	callbacks:Fire("FilterUpdate")
end

function filter:CacheSetsAndUpdate()
	CacheSets()
	-- Should really only fire cache if there were changes
	callbacks:Fire("FilterUpdate")
	callbacks:Fire("PresetUpdate")
end

filter.hooked = filter.hooked or {}
local gHooked = filter.hooked

function HookOutfitterOutfitEvent(pEvent,pTarget)
	if not Outfitter then return end
	if not gHooked[pEvent] then
		Outfitter:RegisterOutfitEvent(pEvent,function(...) filter[pTarget](filter,...) end)
		gHooked[pEvent] = pTarget
	end
end

function filter:Initialize()
	self:CacheSetsAndUpdate()
end

function filter:AddOutfit()
	self:CacheSetsAndUpdate()
end

function filter:DeleteOutfit()
	self:CacheSetsAndUpdate()
end

function filter:EditOutfit()
	Update()
end

local gIsHooked = nil
local gNeedCache = nil

function HookOutfitter()
	if not Outfitter then return end
	if gIsHooked then return end
	gIsHooked = true

	HookOutfitterOutfitEvent("OUTFITTER_INIT","Initialize")
	HookOutfitterOutfitEvent("ADD_OUTFIT","AddOutfit")
	HookOutfitterOutfitEvent("DELETE_OUTFIT","DeleteOutfit")
	HookOutfitterOutfitEvent("EDIT_OUTFIT","EditOutfit")
	gNeedCache = true
	UpdateNextFrame()
	myFrame:UnregisterEvent("ADDON_LOADED")
end

function events.UPDATE()
	myFrame:SetScript("OnUpdate",nil)
	if gNeedCache then
		filter:CacheSetsAndUpdate()
		gNeedCache = nil
	end
	HookOutfitter()
end

function UpdateNextFrame()
	myFrame:SetScript("OnUpdate",function (self, ...) events.UPDATE(...) end)
end

function events.ADDON_LOADED()
	UpdateNextFrame()
end

function events.PLAYER_LOGIN()
	UpdateNextFrame()
	myFrame:RegisterEvent("ADDON_LOADED")
end

local function NextPreset(pUnused, pPreset)
	if pPreset == "" then
		pPreset = nil
	elseif gSetsTable and not pPreset then
		return "", ""
	end
	if gSetsTable then pPreset = next(gSetsTable,pPreset) end
	return pPreset,pPreset and ""
end

function filter:IteratePresets(pSettings)
  return NextPreset, nil, nil
end

function filter:NewSettings(pSettings)
	local vRes
	if pSettings == "" then
		vRes = {
			set = "Any",
		}
	elseif type(pSettings) == "string" then
		vRes = {
			set = pSettings,
		}
	elseif type(pSettings) == "table" then
		vRes = {
			set = pSettings and pSettings.set,
		}
		if pSettings and pSettings.sets then
			local vSets = {}
			for vKey,vValue in pairs(pSettings.sets) do
				vSets[vKey] = vValue
			end
			vRes.sets = vSets
		end
	else
		vRes = {
			set = "Any",
		}
	end
	if not vRes.sets then
		UpgradeSettings(vRes)
	end
	return vRes
end

function filter:FreeSettings(pSettings)
end

function filter:DescribeSettings(pSettings)
	local vSetList
	local vIcon

	if pSettings == "" then
		vSetList = L["Any"]
		vIcon = ICON_OUTFITTER
	elseif type(pSettings) == "string" then
		vSetList = pSettings
		vIcon = gSetIcons and gSetIcons[pSettings] or ICON_QUESTION
	else
		local vSets = pSettings.sets or UpgradeSettings(pSettings)
		if not next(vSets) then
			vSetList = L["None"]
			vIcon = ICON_EMPTY_CRATE
		elseif vSets.Any then
			vSetList = L["Any"]
			vIcon = ICON_OUTFITTER
		else
			local vSetNames = {}
			for vSet in pairs(vSets) do
				tinsert(vSetNames,vSet)
				local vSetIcon = gSetIcons and gSetIcons[vSet] or ICON_QUESTION
				if not vIcon then
					vIcon = vSetIcon
				elseif vIcon ~= vSetIcon then
					vIcon = ICON_WEAPONS
				end
			end
			if not vIcon then
				vIcon = ICON_QUESTION
			end
			tsort(vSetNames)
			vSetList = strjoin(", ",unpack(vSetNames))
		end
	end

	local vDescription = L["Outfitter Set:%s"]:format(vSetList)
	return vDescription,vIcon,vSetList
end

function ToggleSettingsSet(pSettings, pSet)
	if pSet == "Any" then
		pSettings.set = "Any"
		pSettings.sets = { Any = true }
	elseif pSet == "None" then
		pSettings.set = "None"
		pSettings.sets = { }
	else
		local vSets = ExpandSettings(pSettings)

		-- Okay update the given set
		vSets[pSet] = not vSets[pSet] and true or nil

		-- Okay normalize
		vSets.Any = nil -- Since we are expanded, can clear Any
		NormalizeSettings(pSettings)
	end
end

function filter:SettingsOptions(pSettings,pTable,pMethod)
	local vCallback = Update

	if type(pSettings) ~= "table" then return end

	if pTable then
		if type(pTable) == "string" then
			vCallback = function() pTable[pMethod](pTable,pSettings) end
		else
			vCallback = function() pTable(pSettings) end
		end
	end

	local vOptions = {}
	local vMain = {
		type = "group",
		desc = "LibFilter_Outfitter options",
		args = vOptions
	}

	local vCategories = {}
	if gSetsTable then
		if next(gSetsTable) then
			-- Would rather have used a multiselect here -- but not everything supports it
			local vCategoryOrder = Outfitter:GetCategoryOrder()
			for vCategoryIndex, vCategoryID in ipairs(vCategoryOrder) do
				local vOutfits = Outfitter:GetOutfitsByCategoryID(vCategoryID)
				local vCategoryName = Outfitter["c"..vCategoryID.."Outfits"] or vCategoryID
				local vSets = {}
				vCategories[vCategoryName] = {
					type = "group",
					name = vCategoryName,
					desc = vCategoryName,
					args = vSets
				}
				for vIndex, vOutfit in ipairs(vOutfits) do
					if vOutfit and vOutfit.Name then
						local vSet = vOutfit.Name
						vSets[vSet] = {
							type = "toggle",
							name = vSet,
							desc = L["Outfitter Set:%s"]:format(vSet),
							icon = gSetIcons[vSet] or nil,
							get = function(pInfo) return pSettings.sets.Any or pSettings.sets[vSet] end,
							set = function(pInfo, pValue)
											ToggleSettingsSet(pSettings,vSet)
											vCallback()
										end,
						}
					end
				end
			end
		else
			vOptions[1] = {
				order = 1,
				name = L["Outfitter: |cFFFF0000no sets defined|r"],
				desc = L["Outfitter: |cFFFF0000no sets defined|r"],
				type = "header"
			}
			return vMain
		end
	else
		vOptions[1] = {
			order = 1,
			name = L["Outfitter: |cFFFF0000not loaded|r"],
			desc = L["Outfitter: |cFFFF0000not loaded|r"],
			type = "header"
		}
		return vMain
	end

	vOptions[1] = {
		order = 1,
		type = "toggle",
		name = L["Any"],
		desc = L["Any"],
		get = function() return pSettings.sets.Any end,
		set = function(pInfo, pV)
						ToggleSettingsSet(pSettings,"Any")
						vCallback()
					end,
	}
	vOptions[2] = {
		order = 2,
		type = "toggle",
		name = L["None"],
		desc = L["None"],
		get = function() return not next(pSettings.sets) end,
		set = function(pInfo, pV)
						ToggleSettingsSet(pSettings,"None")
						vCallback()
					end,
	}
	vOptions[3] = {
		order = 3,
		name = L["Outfitter Sets"],
		desc = L["Outfitter Sets"],
		type = "group", -- Would prefer this being a multiselect
		args = vCategories,
	}
	return vMain
end

local function TestIfMatchRaw(pSettings,pItemInfo)
	if not pItemInfo then return end
	if not gSetsTable then return end

	if type(pSettings) == "string" then
		if pSettings == "" then
			for vName,vOutfit in pairs(gSetsTable) do
				if Outfitter:OutfitUsesItem(vOutfit, pItemInfo) then return true end
			end
			return nil
		else
			local vOutfit = gSetsTable[pSettings]
			return Outfitter:OutfitUsesItem(vOutfit, pItemInfo)
		end
	else
		local vSets = pSettings.sets or UpgradeSettings(pSettings)
		local vAny = vSets.Any
		local vNone = not next(vSets)

		if vAny then
			for vName,vOutfit in pairs(gSetsTable) do
				if Outfitter:OutfitUsesItem(vOutfit, pItemInfo) then return true end
			end
			return nil
		elseif vNone then
			for vName,vOutfit in pairs(gSetsTable) do
				if Outfitter:OutfitUsesItem(vOutfit, pItemInfo) then return nil end
			end
			return true
		end

		for vSet in pairs(vSets) do
			local vOutfit = gSetsTable[vSet]
			if vOutfit and Outfitter:OutfitUsesItem(vOutfit, pItemInfo) then return true end
		end
	end
	return nil
end

function filter:TestIfMatch(pSettings,pLink)
	if not Outfitter then return end
	if not pLink then return end
	local vItemInfo
	if pLink:match("item:") then
		if not pLink:match("|H") then
			local vName,vLink = GetItemInfo(pLink)
			if vLink then
				pLink = vLink
			else
				pLink = "|cffffffff|H"..pLink.."|h[Unknown]|h"
			end
		end
		vItemInfo = Outfitter:GetItemInfoFromLink(pLink)
	else
		local vBagIndex,vSlotIndex = pLink:match("lot:(-?%d+):(-?%d+)")
		if vBagIndex then
			vItemInfo = Outfitter:GetBagItemInfo(vBagIndex,vSlotIndex)
		end
	end
	return TestIfMatchRaw(pSettings,vItemInfo)
end

function filter:TestIfMatchSplit(pSettings,pType,pBagIndex,pSlotIndex,...)
	if not Outfitter then return end
	local vItemInfo
	if pType == "item" then
		local vString = strjoin(":",pType,pBagIndex,pSlotIndex,...)
		local vName,vLink = GetItemInfo(vString)
		local vItemInfo
		-- Fake itemlink format
		if not vLink then vLink = "|cffffffff|H"..vString.."|h[Unknown]|h" end
		vItemInfo = Outfitter:GetItemInfoFromLink(vLink)
	else
		vItemInfo = Outfitter:GetBagItemInfo(pBagIndex,pSlotIndex)
	end
	return TestIfMatchRaw(pSettings,vItemInfo)
end

myFrame:SetScript("OnEvent",function (self, event, ...) events[event](...) end)

if IsLoggedIn() then
	UpdateNextFrame()
else
	myFrame:RegisterEvent("PLAYER_LOGIN")
end
