--                        -*- Mode: lua; lua-indent-level: 2; tab-width: 2 -*-
--[[
LibFilter_Soulbound

Author: Daphne Pfister <tiiroib02@sneakemail.com>

License

Copyright (c) 2008-2009 Daphne Pfister

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
]]

local MAJOR, MINOR = "LibFilter-item-Soulbound", 303 -- tonumber(("$Revisions$):match("(%d+)"))
local filter = LibStub:NewLibrary(MAJOR, MINOR)
if not filter then return end

-- Avoid Global Lookups
local _G                        = getfenv(0)
local ipairs                    = _G.ipairs
local next                      = _G.next
local pairs                     = _G.pairs
local select                    = _G.select
local string                    = _G.string
local strjoin                   = _G.strjoin
local strsplit                  = _G.strsplit
local table                     = _G.table
local type                      = _G.type
local unpack                    = _G.unpack
local GetContainerItemLink      = _G.GetContainerItemLink
local GetContainerNumSlots      = _G.GetContainerNumSlots
local GetInventoryItemLink      = _G.GetInventoryItemLink
local GetItemCount              = _G.GetItemCount
local GetItemInfo               = _G.GetItemInfo
local KeyRingButtonIDToInvSlotID= _G.KeyRingButtonIDToInvSlotID
local BankButtonIDToInvSlotID   = _G.BankButtonIDToInvSlotID

local ITEM_ACCOUNTBOUND    = _G.ITEM_ACCOUNTBOUND
local ITEM_BIND_TO_ACCOUNT = _G.ITEM_BIND_TO_ACCOUNT
local ITEM_SOULBOUND       = _G.ITEM_SOULBOUND
local ITEM_BIND_ON_PICKUP  = _G.ITEM_BIND_ON_PICKUP
local ITEM_BIND_QUEST      = _G.ITEM_BIND_QUEST
local ITEM_BIND_ON_EQUIP   = _G.ITEM_BIND_ON_EQUIP
local ITEM_BIND_ON_USE     = _G.ITEM_BIND_ON_USE

local KEYRING_CONTAINER    = _G.KEYRING_CONTAINER
local BANK_CONTAINER       = _G.BANK_CONTAINER

local strmatch    = string.match
local tinsert     = table.insert
local tsort       = table.sort

-- Some icons
local ICON_SOULBOUND   = "Interface\\Icons\\Spell_Nature_WispSplode"
local ICON_CHECK       = "Interface\\Buttons\\UI-CheckBox-Check"
local ICON_QUESTION    = "Interface\\Icons\\INV_Misc_QuestionMark"

-- No need for a full blown locale library here
local L
do
	local LL = {
		enUS = {
			["Bound"] = true,
			["Will bind"] = true,
			["Nonbinding"] = true,
			["Matches nothing"] = true,
			["Filter by bind type"] = true,
		},
	}

	local gGameLocale = GetLocale()
	local enL = LL["enUS"]
	L = LL[gGameLocale] or LL["enUS"]

	for vKey,vValue in pairs(enL) do
		if not L[vKey] then L[vKey] = vValue end
	end
	for vKey,vValue in pairs(L) do
		if vValue == true then L[vKey] = vKey end
	end
end

local IS_BOUND = L["Bound"]
local WILL_BIND = L["Will bind"]
local NON_BINDING = L["Nonbinding"]

local BindStates = {
  [ITEM_ACCOUNTBOUND]           = "ITEM_ACCOUNTBOUND",
  [ITEM_BIND_ON_EQUIP]          = "ITEM_BIND_ON_EQUIP",
  [ITEM_BIND_QUEST]             = "ITEM_BIND_QUEST",
  [ITEM_BIND_TO_ACCOUNT]        = "ITEM_BIND_TO_ACCOUNT",
  [ITEM_SOULBOUND]              = "ITEM_SOULBOUND",
  [ITEM_BIND_ON_PICKUP]         = "ITEM_BIND_ON_PICKUP",
  [ITEM_BIND_ON_USE]            = "ITEM_BIND_ON_USE",
}

local CombinedStates = {
  IS_BOUND = {
    ITEM_ACCOUNTBOUND           = true,
    ITEM_SOULBOUND              = true,
    ITEM_BIND_QUEST             = true,
  },
  WILL_BIND = {
    ITEM_BIND_ON_EQUIP          = true,
    ITEM_BIND_TO_ACCOUNT        = true,
    ITEM_BIND_ON_PICKUP         = true,
    ITEM_BIND_ON_USE            = true,
  },
}
local ReverseCombined = {
}
for vGroup,vMembers in pairs(CombinedStates) do
	for vState in pairs(vMembers) do
		ReverseCombined[vState] = ReverseCombined[vState] or {}
		ReverseCombined[vState][vGroup] = true
	end
end


local Descriptions = {
  IS_BOUND                      = IS_BOUND,
  WILL_BIND                     = WILL_BIND,
  NON_BINDING                   = NON_BINDING,
  ITEM_ACCOUNTBOUND             = ITEM_ACCOUNTBOUND,
  ITEM_BIND_ON_EQUIP            = ITEM_BIND_ON_EQUIP,
  ITEM_BIND_QUEST               = ITEM_BIND_QUEST,
  ITEM_BIND_TO_ACCOUNT          = ITEM_BIND_TO_ACCOUNT,
  ITEM_SOULBOUND                = ITEM_SOULBOUND,
  ITEM_BIND_ON_PICKUP           = ITEM_BIND_ON_PICKUP,
  ITEM_BIND_ON_USE              = ITEM_BIND_ON_USE,
}

local presets = {
  IS_BOUND                      = "",
  WILL_BIND                     = "",
  NON_BINDING                   = "",
  ITEM_ACCOUNTBOUND             = "",
  ITEM_BIND_ON_EQUIP            = "",
  ITEM_BIND_QUEST               = "",
  ITEM_BIND_TO_ACCOUNT          = "",
  ITEM_SOULBOUND                = "",
  ITEM_BIND_ON_PICKUP           = "",
  ITEM_BIND_ON_USE              = "",
}

filter.author = "Nayala"
filter.credits = ""
filter.description = L["Filter by bind type"]
filter.filtertype = "item"
filter.filtertypes = filter.filtertypes or {}
filter.filtertypes["item"] = 1
filter.filtertypes["Slot"] = 1
filter.filtertypes["Inventory"] = 1

filter.key = "Soulbound"
filter.name = ITEM_SOULBOUND
filter.icon = ICON_SOULBOUND

filter.callbacks = filter.callbacks or LibStub("CallbackHandler-1.0"):New(filter)
filter.events = filter.events or {}
filter.myFrame = filter.myFrame or CreateFrame("Frame","LibFilter_Soulbound_Frame")
local callbacks = filter.callbacks
local events = filter.events
local myFrame = filter.myFrame
local tooltip, llines, rlines

filter.extensions = filter.extensions or {}
local extensions = filter.extensions
extensions.Baggins = extensions.Baggins or {}
--extensions.Baggins.name = "Bind"

-- Prototypes
local Print -- (msg)
local Update
local CreateTooltip
local EraseTooltip
local DeHyperlink -- (link)
local GetLinkType -- (linkortype)
local GetLinkArgs -- (linkortype, ...)
local SetBagSlot -- (bag,slot)
local GuessBagSlot -- (link)
local GetBoundStatus -- (linkortype, ...)
local NormalizeState -- (settings)
local ToggleState -- (settings, state)

function Print(msg)
	local info = ChatTypeInfo["SYSTEM"]
	DEFAULT_CHAT_FRAME:AddMessage(msg, info.r, info.g, info.b, info.id)
end

function events.BANKFRAME_OPENED()
	filter.bankOpen = true
end

function events.BANKFRAME_CLOSED()
	filter.bankOpen = false
end

function Update()
	callbacks:Fire("FilterUpdate")
end

function CreateTooltip()
	if tooltip then return end
	tooltip = filter.tooltip or CreateFrame("GameTooltip")
	llines = filter.llines or {}
	rlines = filter.rlines or {}
	filter.tooltip, filter.llines, filter.rlines = tooltip, llines, rlines

	tooltip:SetOwner(UIParent, "ANCHOR_NONE")
	for i=1,30 do
		if not llines[i] then
			local left,right = tooltip:CreateFontString(), tooltip:CreateFontString()
			llines[i],rlines[i] = left,right
			left:SetFontObject(GameFontNormal)
			right:SetFontObject(GameFontNormal)
			tooltip:AddFontStrings(left,right)
		end
	end
end

function EraseTooltip()
	if not tooltip then CreateTooltip() end
	tooltip:ClearLines()
	for i=1,30 do
		--llines[i]:SetTextColor(0,0,0)
		rlines[i]:SetText()
		--rlines[i]:SetTextColor(0,0,0)
	end
	if not tooltip:IsOwned(UIParent) then
		tooltip:SetOwner(UIParent, "ANCHOR_NONE")
	end
end

function DeHyperlink(link)
	return strmatch(link or "","|H([^|]+)|") or link
end

function GetLinkType(linkortype)
	return strsplit(":",linkortype or "",2)
end

function GetLinkArgs(type,linkortype,...)
	if linkortype == type then
		return type,...
	else
		return strsplit(":",linkortype)
	end
end

function SetBagSlot(bag,slot)
	if bag == -99 then
		tooltip:SetInventoryItem("player",slot)
	elseif bag == KEYRING_CONTAINER then
		tooltip:SetInventoryItem("player",KeyRingButtonIDToInvSlotID(slot))
	elseif bag == BANK_CONTAINER then
		tooltip:SetInventoryItem("player",BankButtonIDToInvSlotID(slot))
	else
		tooltip:SetBagItem(bag,slot)
	end
end

function GuessBagSlot(link)
	if not link then return nil, nil end
	if GetItemCount(link) < 1 then return nil, nil end

	local _,reallink = GetItemInfo(link)
	if not reallink then reallink = link end

	-- Try bags
	local highBag = NUM_BAG_SLOTS
	if filter.bankOpen then highBag = highBag + NUM_BANKBAGSLOTS end
	for vBag = -2,highBag do
		if filter.bankOpen or vBag ~= -1 then
			for vSlot = 1,GetContainerNumSlots(vBag) do
				if GetContainerItemLink(vBag,vSlot) == link then
					return vBag,vSlot
				end
			end
		end
	end

	-- Try inventory
	-- TODO: Clean up magic numbers
	for vSlot = 0,23 do
		if GetInventoryItemLink("player",vSlot) == link then
			return -99,vSlot
		end
	end

	if filter.bankOpen then
		for vSlot = 68,74 do
			if GetInventoryItemLink("player",vSlot) == link then
				return -99,vSlot
			end
		end
	end

	return nil,nil
end

function GetBoundStatus(linkortype, ...)
	if not linkortype then return "NON_BINDING" end
	if not tooltip then CreateTooltip() end
	EraseTooltip()
	linkortype = DeHyperlink(linkortype)
	local type = GetLinkType(linkortype)
	local needfix
	if type == "Slot" then
		local _,bag,slot = GetLinkArgs(type,linkortype,...)
		SetBagSlot(tonumber(bag),tonumber(slot))
	elseif type == "Inventory" then
		local _,unit,slot = GetLinkArgs(type,linkortype,...)
		if unit == "" then unit = "player" end
		tooltip:SetInventoryItem(unit,tonumber(slot))
	elseif type == "item" then
		if type == linkortype then
			linkortype = strjoin(":",linkortype,...)
		end
		local bag,slot = GuessBagSlot(linkortype)
		if bag and slot then
			SetBagSlot(bag,slot)
		else
			tooltip:SetHyperlink(linkortype)
		end
	else
		return "NON_BINDING"
	end
	local lines = tooltip:NumLines()
	for i=2,2 do
		local text = llines[i]:GetText()
		if text and BindStates[text] then
			return BindStates[text]
		end
	end
	return "NON_BINDING"
end

function NormalizeState(settings)
	-- Fix combinations
	for combination, members in pairs(CombinedStates) do
		local anyclear = false
		for state in pairs(members) do
			if not settings[state] then
				anyclear = true
			end
		end
		settings[combination] = not anyclear
	end
end

function ToggleState(settings, togglestate)
	local members = CombinedStates[togglestate]
	if members then
		local anyclear = false
		for state in pairs(members) do
			if not settings[state] then
				anyclear = true
			end
		end
		-- If anyclear then should set all, otherwise clear all
		for state in pairs(members) do
			settings[state] = anyclear
		end
	elseif Descriptions[togglestate] then
		settings[togglestate] = not settings[togglestate]
	end
	NormalizeState(settings)
end

function filter:IteratePresets(pSettings)
  return pairs(presets)
end

function filter:NewSettings(pSettings)
	local vRes = {}
	if pSettings == "" then
		ToggleState(vRes,"IS_BOUND")
	elseif type(pSettings) == "string" then
		ToggleState(vRes,pSettings)
	elseif type(pSettings) == "table" then
		for state, value in pairs(pSettings) do
			vRes[state] = value
		end
		NormalizeState(vRes)
	else
		ToggleState(vRes,"IS_BOUND")
	end
	return vRes
end

function filter:FreeSettings(pSettings)
end

function filter:DescribeSettings(pSettings)
	local vStateList
	local vIcon = ICON_SOULBOUND

	if pSettings == "" then
		vStateList = IS_BOUND
	elseif type(pSettings) == "string" then
		vStateList = Descriptions[pSettings] or IS_BOUND
	else
		-- TODO: Have filtername part of long name
		local vStateNames = {}
		for vState, vValue in pairs(pSettings) do
			local vCheckGroups = ReverseCombined[vState]
			local vOk = vValue
			if vCheckGroups then
				for vGroup in pairs(vCheckGroups) do
					if pSettings[vGroup] then vOk = false
					end
				end
			end
			if vOk and Descriptions[vState] then
				tinsert(vStateNames,Descriptions[vState])
			end
		end
		tsort(vStateNames)
		vStateList = strjoin(", ",unpack(vStateNames))
	end

	return vStateList,vIcon,vStateList
end

function filter:SettingsOptions(pSettings,pTable,pMethod)
	local vCallback = Update

	if type(pSettings) ~= "table" then return end

	if pTable then
		if type(pTable) == "string" then
			vCallback = function() pTable[pMethod](pTable,pSettings) end
		else
			vCallback = function() pTable(pSettings) end
		end
	end

	local vOptions = {}
	local vOptionProto = {}
	local vMain = {
		type = "group",
		desc = "LibFilter_Soulbound options",
		args = vOptions
	}

	for vName, vLocalized in pairs(Descriptions) do
		vOptionProto[vName] = {
			type = "toggle",
			name = vLocalized,
			desc = vLocalized,
			get = function(pInfo) return pSettings[vName] end,
			set = function(pInfo)
							ToggleState(pSettings,vName)
							vCallback()
						end,
		}
	end
	for vName, vLocalized in pairs(Descriptions) do
		if CombinedStates[vName] then
			local vGroup = { vOptionProto[vName] }
			local vHead = {
				type = "group",
				name = vLocalized,
				desc = vLocalized,
				args = vGroup,
			}
			vOptions["GROUP:"..vName] = vHead
			vGroup[1] = vOptionProto[vName]
			vGroup[1].order = 1
			vGroup[2] = {
				order = 2,
				type = "header",
			}
			local order = 3
			local anyset
			for vMember in pairs(CombinedStates[vName]) do
				order = order + 1
				vGroup[order] = vOptionProto[vMember]
				vGroup[order].order = order
				if pSettings[vMember] then anyset = true end
			end
			if pSettings[vName] then
				vHead.icon = ICON_CHECK
			elseif anyset then
				vHead.icon = ICON_QUESTION
			end
		elseif ReverseCombined[vName] then
		else
			vOptions[vName] = vOptionProto[vName]
		end
	end
	return vMain
end

function filter:TestIfMatch(settings,...)
	local bindtype = GetBoundStatus(...)

	if type(settings) == "string" then
		if settings == "" then settings = "IS_BOUND" end
		if bindtype == settings then return true end
		-- check group
		local members = CombinedStates[settings]
		if members and members[bindtype] then return true end
		return false
	else
		return settings[bindtype] and true or false
	end
	return nil
end

filter.TestIfMatchSplit = filter.TestIfMatch

myFrame:SetScript("OnEvent",function (self, event, ...) events[event](...) end)
myFrame:RegisterEvent("BANKFRAME_OPENED")
myFrame:RegisterEvent("BANKFRAME_CLOSED")
